@@ -1,5 +1,5 @@
 #!/usr/bin/env python2
-#=============================================================================
+#
 # Copyright 2016 by Shaheed Haque (srhaque@theiet.org)
 # Copyright 2016 by Stephen Kelly (steveire@gmail.com)
 #
@@ -25,7 +25,7 @@
 # THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 # THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-#=============================================================================
+#
 
 """SIP file generator for PyQt."""
 
@@ -36,7 +36,6 @@ import inspect
 import logging
 import os
 import re
-import subprocess
 import sys
 import traceback
 from clang import cindex
@@ -62,6 +61,35 @@ TEMPLATE_KINDS = [
                      CursorKind.TYPE_REF, CursorKind.TEMPLATE_REF, CursorKind.NAMESPACE_REF
                  ] + EXPR_KINDS
 
+def clang_diagnostic_to_logging_diagnostic(lvl):
+    """
+
+    The diagnostic levels in cindex.py are
+
+        Ignored = 0
+        Note    = 1
+        Warning = 2
+        Error   = 3
+        Fatal   = 4
+
+    and the leves in the python logging module are
+
+        NOTSET      0
+        DEBUG       10
+        INFO        20
+        WARNING     30
+        ERROR       40
+        CRITICAL    50
+
+    """
+    return (logging.NOTSET,
+        logging.INFO,
+        logging.WARNING,
+        logging.ERROR,
+        logging.CRITICAL)[lvl]
+
+def diagnostic_word(lvl):
+    return ("", "info", "warning", "error", "fatality")[lvl]
 
 class SipGenerator(object):
     def __init__(self, project_rules, compile_flags, verbose=False, dump_includes=False, dump_privates=False):
@@ -114,10 +142,13 @@ class SipGenerator(object):
             #
             loc = diag.location
             msg = "{}:{}[{}] {}".format(loc.file, loc.line, loc.column, diag.spelling)
+            if (diag.spelling == "#pragma once in main file"):
+                continue
             if msg in self.diagnostics:
                 continue
             self.diagnostics.add(msg)
-            logger.log(diag.severity, "Parse error {}".format(msg))
+            logger.log(clang_diagnostic_to_logging_diagnostic(diag.severity),
+                "Parse {}: {}".format(diagnostic_word(diag.severity), msg))
         if self.dump_includes:
             for include in sorted(set(self.tu.get_includes())):
                 logger.debug(_("Used includes {}").format(include.include.name))
@@ -127,10 +158,21 @@ class SipGenerator(object):
         body = self._container_get(self.tu.cursor, -1, h_file, include_filename)
         return body, self.tu.get_includes
 
-    CONTAINER_SKIPPABLE_UNEXPOSED_DECL = re.compile("_DECLARE_PRIVATE|friend|;")
-    FN_SKIPPABLE_ATTR = re.compile("_EXPORT|Q_REQUIRED_RESULT|format\(printf")
-    VAR_SKIPPABLE_ATTR = re.compile("_EXPORT")
-    TYPEDEF_SKIPPABLE_ATTR = re.compile("_EXPORT")
+    def skippable_attribute(self, parent, member, text, sip):
+        """
+        We don't seem to have access to the __attribute__(())s, but at least we can look for stuff we care about.
+
+        :param member:          The attribute.
+        :param text:            The raw source corresponding to the region of member.
+        """
+        if member.kind != CursorKind.VISIBILITY_ATTR:
+            return False
+        if member.spelling == "hidden":
+            if self.dump_privates:
+                logger.debug("Ignoring private {}".format(SipGenerator.describe(parent)))
+            sip["name"] = ""
+            return True
+        return False
 
     def _container_get(self, container, level, h_file, include_filename):
         """
@@ -142,17 +184,19 @@ class SipGenerator(object):
         :return:                    A string.
         """
 
-        def skippable_attribute(member, text):
-            """
-            We don't seem to have access to the __attribute__(())s, but at least we can look for stuff we care about.
-
-            :param member:          The attribute.
-            :param text:            The raw source corresponding to the region of member.
-            """
-            if text.find("_DEPRECATED") != -1:
-                sip["annotations"].add("Deprecated")
-                return True
-            SipGenerator._report_ignoring(container, member, text)
+        if container.kind.is_translation_unit():
+            #
+            # Any module-related manual code (%ExportedHeaderCode, %ModuleCode, %ModuleHeaderCode or other
+            # module-level directives?
+            #
+            sip = {
+                "name": include_filename,
+                "decl": ""
+            }
+            self.rules.modulecode(include_filename, sip)
+            body = sip["code"]
+        else:
+            body = ""
 
         sip = {
             "name": container.displayname,
@@ -163,7 +207,6 @@ class SipGenerator(object):
             if self.dump_privates:
                 logger.debug("Ignoring private {}".format(SipGenerator.describe(container)))
             return ""
-        body = ""
         base_specifiers = []
         template_type_parameters = []
         had_copy_constructor = False
@@ -183,7 +226,7 @@ class SipGenerator(object):
             elif member.kind == CursorKind.CXX_ACCESS_SPEC_DECL:
                 decl = self._get_access_specifier(member, level + 1)
             elif member.kind == CursorKind.TYPEDEF_DECL:
-                decl = self._typedef_get(member, level + 1)
+                decl = self._typedef_get(container, member, level + 1)
             elif member.kind == CursorKind.CXX_BASE_SPECIFIER:
                 #
                 # Strip off the leading "class". Except for TypeKind.UNEXPOSED...
@@ -211,7 +254,12 @@ class SipGenerator(object):
                 #
                 pass
             else:
-                SipGenerator._report_ignoring(container, member)
+                text = self._read_source(member.extent)
+                if self.skippable_attribute(container, member, text, sip):
+                    if not sip["name"]:
+                        return ""
+                else:
+                    SipGenerator._report_ignoring(container, member)
 
             def is_copy_constructor(member):
                 if member.kind != CursorKind.CONSTRUCTOR:
@@ -268,30 +316,51 @@ class SipGenerator(object):
                     pad = " " * ((level + 1) * 4)
                     body += pad + "// {}\n".format(SipGenerator.describe(member))
                 body += decl
+
+
+        if container.kind == CursorKind.TRANSLATION_UNIT:
+            return body
+
+        if container.kind == CursorKind.NAMESPACE:
+            container_type = "namespace " + name
+        elif container.kind in [CursorKind.CLASS_DECL, CursorKind.CLASS_TEMPLATE,
+                                CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION]:
+            container_type = "class " + name
+        elif container.kind == CursorKind.STRUCT_DECL:
+            container_type = "struct " + name
+        elif container.kind == CursorKind.UNION_DECL:
+            container_type = "union " + name
+        else:
+            raise AssertionError(
+                _("Unexpected container {}: {}[{}]").format(container.kind, name, container.extent.start.line))
+
+        sip["decl"] = container_type
+        sip["template_parameters"] = template_type_parameters
+
+        pad = " " * (level * 4)
+
         #
         # Empty containers are still useful if they provide namespaces or forward declarations.
         #
-        if not body and level >= 0:
-            body = "\n"
+        if not body:
             text = self._read_source(container.extent)
             if not text.endswith("}"):
                 #
                 # Forward declaration.
                 #
-                sip["annotations"].add("External")
-        if body and level >= 0:
-            if container.kind == CursorKind.NAMESPACE:
-                container_type = "namespace " + name
-            elif container.kind in [CursorKind.CLASS_DECL, CursorKind.CLASS_TEMPLATE,
-                                    CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION]:
-                container_type = "class " + name
-            elif container.kind == CursorKind.STRUCT_DECL:
-                container_type = "struct " + name
-            elif container.kind == CursorKind.UNION_DECL:
-                container_type = "union " + name
-            else:
-                raise AssertionError(
-                    _("Unexpected container {}: {}[{}]").format(container.kind, name, container.extent.start.line))
+                modifying_rule = self.rules.forward_declaration_rules().apply(container, sip)
+                if sip["name"]:
+                    if modifying_rule:
+                        body += "// Modified {} (by {}):\n".format(SipGenerator.describe(container), modifying_rule)
+                    if "External" in sip["annotations"]:
+                        body += pad + sip["decl"]
+                        body += " /External/;\n"
+                    else:
+                        body = pad + "// Discarded {} (by {})\n".format(SipGenerator.describe(container), "default forward declaration handling")
+
+                else:
+                    body = pad + "// Discarded {} (by {})\n".format(SipGenerator.describe(container), modifying_rule)
+        else:
             #
             # Generate private copy constructor for non-copyable types.
             #
@@ -300,33 +369,24 @@ class SipGenerator(object):
             #
             # Flesh out the SIP context for the rules engine.
             #
-            sip["template_parameters"] = ", ".join(template_type_parameters)
-            sip["decl"] = container_type
-            sip["base_specifiers"] = ", ".join(base_specifiers)
+            sip["base_specifiers"] = base_specifiers
             sip["body"] = body
             modifying_rule = self.rules.container_rules().apply(container, sip)
-            pad = " " * (level * 4)
             if sip["name"]:
                 decl = ""
                 if modifying_rule:
                     decl += "// Modified {} (by {}):\n".format(SipGenerator.describe(container), modifying_rule)
                 decl += pad + sip["decl"]
-                if "External" in sip["annotations"]:
-                    #
-                    # SIP /External/ does not seem to work as one might wish. Suppress.
-                    #
-                    body = decl + " /External/;\n"
-                    body = pad + "// Discarded {} (by {})\n".format(SipGenerator.describe(container), "/External/ handling")
-                else:
-                    if sip["base_specifiers"]:
-                        decl += ": " + sip["base_specifiers"]
-                    if sip["annotations"]:
-                        decl += " /" + ",".join(sip["annotations"]) + "/"
-                    if sip["template_parameters"]:
-                        decl = pad + "template <" + sip["template_parameters"] + ">\n" + decl
-                    decl += "\n" + pad + "{\n"
-                    decl += "%TypeHeaderCode\n#include <{}>\n%End\n".format(include_filename)
-                    body = decl + sip["body"] + pad + "};\n"
+
+                if sip["base_specifiers"]:
+                    decl += ": " + ", ".join(sip["base_specifiers"])
+                if sip["annotations"]:
+                    decl += " /" + ",".join(sip["annotations"]) + "/"
+                if sip["template_parameters"]:
+                    decl = pad + "template <" + ", ".join(sip["template_parameters"]) + ">\n" + decl
+                decl += "\n" + pad + "{\n"
+                decl += "%TypeHeaderCode\n#include <{}>\n%End\n".format(include_filename)
+                body = decl + sip["body"] + pad + "};\n"
             else:
                 body = pad + "// Discarded {} (by {})\n".format(SipGenerator.describe(container), modifying_rule)
         return body
@@ -335,10 +395,22 @@ class SipGenerator(object):
         """
         Skip access specifiers embedded in the Q_OBJECT macro.
         """
-        access_specifier = self._read_source(member.extent)
-        if access_specifier == "Q_OBJECT":
+        access_specifier_text = self._read_source(member.extent)
+        if access_specifier_text == "Q_OBJECT":
             return ""
         pad = " " * ((level - 1) * 4)
+        access_specifier = ""
+        if (access_specifier_text in ("Q_SIGNALS:", "signals:",
+                                      "public Q_SLOTS:", "public slots:",
+                                      "protected Q_SLOTS:", "protected slots:")):
+            access_specifier = access_specifier_text
+        elif member.access_specifier == AccessSpecifier.PRIVATE:
+            access_specifier = "private:"
+        elif member.access_specifier == AccessSpecifier.PROTECTED:
+            access_specifier = "protected:"
+        elif member.access_specifier == AccessSpecifier.PUBLIC:
+            access_specifier = "public:"
+
         decl = pad + access_specifier + "\n"
         return decl
 
@@ -365,17 +437,12 @@ class SipGenerator(object):
         :param level:               Recursion level controls indentation.
         :return:                    A string.
         """
-
-        def skippable_attribute(member, text):
-            """
-            We don't seem to have access to the __attribute__(())s, but at least we can look for stuff we care about.
-
-            :param member:          The attribute.
-            :param text:            The raw source corresponding to the region of member.
-            """
-            if SipGenerator.FN_SKIPPABLE_ATTR.search(text):
-                return True
-            SipGenerator._report_ignoring(function, member, text)
+        if container.kind == CursorKind.TRANSLATION_UNIT and \
+                (function.semantic_parent.kind == CursorKind.CLASS_DECL or
+                 function.semantic_parent.kind == CursorKind.STRUCT_DECL) and \
+                function.is_definition():
+            # Skip inline methods
+            return
 
         sip = {
             "name": function.spelling,
@@ -424,9 +491,9 @@ class SipGenerator(object):
                 template_parameters.append(child.type.spelling + " " + child.displayname)
             else:
                 text = self._read_source(child.extent)
-                if child.kind in [CursorKind.UNEXPOSED_ATTR, CursorKind.VISIBILITY_ATTR] and skippable_attribute(child,
-                                                                                                                 text):
-                    pass
+                if self.skippable_attribute(function, child, text, sip):
+                    if not sip["name"]:
+                        return ""
                 else:
                     SipGenerator._report_ignoring(function, child)
         #
@@ -442,6 +509,11 @@ class SipGenerator(object):
         modifying_rule = self.rules.function_rules().apply(container, function, sip)
         pad = " " * (level * 4)
         if sip["name"]:
+            #
+            # Any method-related code (%MethodCode, %VirtualCatcherCode, VirtualCallCode
+            # or other method-related directives)?
+            #
+            self.rules.methodcode(function, sip)
             decl = ""
             if modifying_rule:
                 decl += "// Modified {} (by {}):\n".format(SipGenerator.describe(function), modifying_rule) + pad
@@ -449,14 +521,14 @@ class SipGenerator(object):
             if parameter_modifying_rules:
                 decl += pad
 
-            sip["template_parameters"] = ", ".join(sip["template_parameters"])
             decl += sip["name"] + "(" + ", ".join(sip["parameters"]) + ")"
             if sip["fn_result"]:
                 decl = sip["fn_result"] + " " + decl
             decl = pad + sip["prefix"] + decl + sip["suffix"]
             if sip["template_parameters"]:
-                decl = pad + "template <" + sip["template_parameters"] + ">\n" + decl
+                decl = pad + "template <" + ", ".join(sip["template_parameters"]) + ">\n" + decl
             decl += ";\n"
+            decl += sip["code"]
         else:
             decl = pad + "// Discarded {} (by {})\n".format(SipGenerator.describe(function), modifying_rule)
         return decl
@@ -498,6 +570,14 @@ class SipGenerator(object):
         def _get_param_type(parameter):
             result = parameter.type.get_declaration().type
 
+            if result.kind != TypeKind.ENUM and result.kind != TypeKind.TYPEDEF and parameter.type.kind == TypeKind.LVALUEREFERENCE:
+                if parameter.type.get_pointee().get_declaration().type.kind != TypeKind.INVALID:
+                    return parameter.type.get_pointee().get_declaration().type
+                return parameter.type.get_pointee()
+
+            if parameter.type.get_declaration().type.kind == TypeKind.INVALID:
+                return parameter.type
+
             if (parameter.type.get_declaration().type.kind == TypeKind.TYPEDEF):
                 isQFlags = False
                 for member in parameter.type.get_declaration().get_children():
@@ -510,8 +590,16 @@ class SipGenerator(object):
             return result
 
         def _get_param_value(text, parameterType):
-            if text == "0":
+            if text == "0" or text == "nullptr":
                 return text
+            if text == "{}":
+                if parameterType.kind == TypeKind.ENUM:
+                    return "0"
+                if parameterType.kind == TypeKind.POINTER:
+                    return "nullptr"
+                if parameterType.spelling.startswith("const "):
+                    return parameterType.spelling[6:] + "()"
+                return parameterType.spelling + "()"
             if not "::" in parameterType.spelling:
                 return text
             try:
@@ -564,7 +652,7 @@ class SipGenerator(object):
                 return _get_param_value(text, parameterType)
         return ""
 
-    def _typedef_get(self, typedef, level):
+    def _typedef_get(self, container, typedef, level):
         """
         Generate the translation for a typedef.
 
@@ -574,10 +662,20 @@ class SipGenerator(object):
         :return:                    A string.
         """
 
-        pad = " " * (level * 4)
+        sip = {
+            "name": typedef.displayname,
+            "decl": typedef.underlying_typedef_type.spelling,
+            "annotations": set(),
+        }
 
-        decl = pad + "typedef {} {}".format(typedef.underlying_typedef_type.spelling, typedef.displayname)
-        decl += ";\n"
+        self.rules.typedef_rules().apply(container, typedef, sip)
+
+        pad = " " * (level * 4)
+        if sip["name"]:
+            decl = pad + "typedef {} {}".format(sip["decl"], sip["name"])
+            decl += ";\n"
+        else:
+            decl = pad + "// Discarded {}\n".format(SipGenerator.describe(typedef))
         return decl
 
     def _var_get(self, container, variable, level):
@@ -590,17 +688,6 @@ class SipGenerator(object):
         :return:                    A string.
         """
 
-        def skippable_attribute(member, text):
-            """
-            We don't seem to have access to the __attribute__(())s, but at least we can look for stuff we care about.
-
-            :param member:          The attribute.
-            :param text:            The raw source corresponding to the region of member.
-            """
-            if SipGenerator.VAR_SKIPPABLE_ATTR.search(text):
-                return True
-            SipGenerator._report_ignoring(container, member, text)
-
         sip = {
             "name": variable.spelling
         }
@@ -614,8 +701,9 @@ class SipGenerator(object):
                 pass
             else:
                 text = self._read_source(child.extent)
-                if child.kind == CursorKind.VISIBILITY_ATTR and skippable_attribute(child, text):
-                    pass
+                if self.skippable_attribute(variable, child, text, sip):
+                    if not sip["name"]:
+                        return ""
                 else:
                     SipGenerator._report_ignoring(variable, child)
         #
